# Spook CTF evaluation scripts

This directory contains the scripts that can be used to evaluate a submission.
They are designed to be maximally portable on linux machines. You should run
your submission package through these scripts to ensure that all the file
formats and scripts are correct.


## Setup & Dependencies

Dependencies:
- linux/unix environment (including `bash`)
- `singularity 3.5`: Container tool for HPC <https://sylabs.io/>.

Run
```
setup.sh
```
to initialize the singularity containers used by the scripts.


## Run

The top-level script is `main.sh`.
It takes as input a submission zip file, a directory containing the training
traces and a directory containing the challenge traces.

The evaluation is done in a few stages:
- unpack:
  The submission zip file is unpacked and the json description file is validated.
- setup:
  The setup script from the submission is run.
- profile:
  The profile script from the submission for each of the targets is run.
- attack:
  The attack script from the submission for each of the targets is run,
  for each of the combinations of number of traces and key id.
- rank:
  The rank of the correct key is computed from the attack results.

Those stages can be run using
```
main.sh [STAGE]
```
or `scritps/main.sh all` to run all the stages.

### Environment variables

The `main.sh` script is configured through the following environment variables:
- `CTF_SUB_FILE`: Submission zip file.
- `CTF_BASE_DIR`: Scratch directory containing all the temporary files of the evaluation.
- `CTF_TRACES_TRAIN_DIR`:
  Top-level directory containing the training sets (shall be set to the
  top-level public set during the envaluation, that is, the directory
  containing sw3/, sw4/, etc.)
- `CTF_CHALLENGE_DIR`:
  Top-level directory containing the traces to attack. Shall have identical
  structure to the `CTF_TRACES_TRAIN_DIR`.
  The user running this script must have read and write access to all the
  files in this directory.
  This can be set to the same value of `CTRF_TRACES_TRAIN_DIR` to evaluate the
  attack on the training dataset.
- `CTF_CHALLENGE_TMP`:
  Temporary directory to store temporary challenge traces. This directory
  contains hardlinks to files in `CTF_CHALLENGE_DIR`, hence must be on the same
  partition.
- `CTF_N_ATTACK_SETS`:
  Number of different keys on which the attack should be run (goes from `key_0`
  to `key_[x-1]` when this value is set to `x`.
- `CTF_SUDO_RESOURCES`:
  Set to 1 to limit resources (CPU, RAM) usage according to the challenge
  rules. Requires sudo rights.


## Files and directories format

### Submission Zip

The submission zip shall contain the following files:
- `/submission.json`: the submission description file (see below)
- `/setup`: executable file run a root on a fresh Debian Buster container OS.
  The OS will subsequently be used to run the profiling and the attacks.
  The execution timeout is 20 minutes with 8 GB of RAM and 8 CPU cores.
- `/[TARGET]/{profile,attack}` executable files for each of the exploited targets.

### Submission description file (submission.json)

The submission description file shall conform to the json standard and contain
an object containing the following fields:
- `"authors"`: the list of authors, in the form of an array of object, each
  containing the `"name"` field (that may contain a pseudonym) and the
  `"email"` field (there must at least one valid email address).
- `"name"`: a name for the submission. This should remain the same for the all
  the submissions made at different contest deadlines by the same authors.
- `"license"`: the [SPDX identifier](https://spdx.org/licenses/) of an
  OSI-approved license under which the work can be distributed.
  The submitter distributes the submission to the constest organisers under the
  terms of the license disignated by this identifier.
- `"attacks"`: an object whose keys are the targets that the submission
  exploits (a sub-set of `sw3`, `sw4`, `sw6`, `sw8`, `hw2`, `hw3`, `hw4`) and
  whose values are object whose unique key is `"n_traces"`, the value attached
  to which is an array of at most 8 numbers.
  Those numbers designate the number of traces on which the attack should be
  tested (see the [contest rules](https://ctf.spook.dev/rules.html)
  for details).
  
### Profile script

The intended use-case for the profile script is to profile the attack
using the supplied traces. In case no profiling is needed (or its result is
submitted in the submission package), this script can be an empty shell script.
All the training traces are available read-only at `/profile_traces`.
The `/profile_scratch` directory is mounted read-write (intended to store the
result of the profiling).

### Attack script

All the training traces are available read-only at `/profile_traces`.
The traces to be attacked are stored in a series of `.npz` files in the
`/challenge_traces` directory.
The attack results should be printed on the standard output.

The attack script shall output (on stdin and stderr combined) lines that match the following format:
- either comments starting with `#`,
- either key sub-part likelihoods.

A key-sub-part likelihood line has the following format
```
MSK:L0,L1,L2...,L255
```
Where MSK is a 128-bit HEX bitmask with 8 bits set to 1, L0 (resp LX) is the
log-likelihood that the concatenation of the bits selected by the mask is
equal to 0 (resp. X).
There must thus be 16 key sub-part likelihood lines, to cover all the 128 bits,
and each line must cover distinct bits.

### Training traces

The training traces dataset has the following structure
```
[TARGET]/
    fixed_key/
        key_[X]/
            fkey_[TARGET]_K[X]_[N]_[Y].npz
            secret_[TARGET]_K[X].txt
    random_key/
        rkey_[TARGET]_[N]_[Y].npz
```
where
- `TARGET` is one of `sw3`, `sw4`, `sw6`, `sw8`, `hw2`, `hw3`, `hw4`
- `X` is between 0 and 4
- `N` is the number of traces in the file
- `Y` is a counter

Two types of datasets are provided. The files under `random_keys/`
contain traces with random inputs (typically used for profiling). 
The files under `fixed_key/key_[X]` contain traces with the same 
key with one different key per value of X (typically used to evaluate the attacks). 

For software targets with random keys, the files contain the following fields:
- `traces`: is a int16 matrix where each of the nt rows is a trace corresponding to an encryption. Each file contains nt=10k traces.
- `nonces`: is a (nt,4) uint32 matrix where each row is the nonce used for the corresponding encryption.
- `umsk_keys`: is a (nt,4) uint32 matrix  where each row is the secret key.
- `msk_keys`: is a (nt,4\*D) uint32 matrix  where each row is the D sharing of the secret key used. The shares are layout as described above.
- `seeds`: is a (nt,4) uint32 matrix where each row is a 128-bit word used to seed the prng.
- `ad`: is the associated data. It is fixed to zero since it has no impact on the first Clyde call.
- `m`:is the message. It is fixed to zero since it has no impact on the first Clyde call.

For the hardware targets with random keys, the files contain the following fields:
- `ram_in`: array containing the successive 32-bits commands sent to the HW core for each case. These 
are stored as a long bytestrings. 
- `msk_keys`: array containing the (masked) keys used. These are stored as d\*16 bytes long 
bytestrings. 
- `umsk_keys`: array containing the (unmasked) keys used. These are stored as 16 bytes long 
bytestrings.
- `s0`: array containing the seed used to generate the key refresh randomness. These are stored as 16 
bytes long bytestrings.
- `s1`: array containing the seed used to generate the first randomness of the sbox layer. These are stored as 16 
bytes long bytestrings.
- `s2`: array containing the seed used to generate the second randomness of the sbox layer. These are stored as 16 
bytes long bytestrings.
- `nonce`: array containing the nonce used. These are stored as 16 bytes long bytestrings.
- `traces`: array containing the traces measured. These are stored as vector of np.int16.

For the fixed keys files (both sw and hw), the only field provided are `traces` and `nonces`.


### Challenge traces

The challenge traces are contained in multiple `.npz` files, all relating to
the same key. The format of the file is identical to the `fkey_` files of the
corresponding training dataset.

### Ranking result

The key ranking results are in the files `$CTF_BASE_DIR/ranking/[TARGET]`, where each line is of the form:
```
NUMBER_OF_TRACES LB_KEY0,ESTIMATE_KEY0,UP_KEY0;LB_KEY1,ESTIMATE_KEY1,UP_KEY1;...
```
The `ESTIMATE` values are the log of the key rank.
The final rank to the given number of traces is the average of the key ranks
(i.e., average of `2^ESTIMATE`).


## More details

More information is available in the header comments of the individual scripts.

## Bugs

Your are welcome to report any bug you may find in this package to Gaëtan
Cassiers <gaetan.cassiers@uclouvain.be>.

## License

The spook CTF evaluation scripts are published under the MIT license. See
<COPYRIGHT> for details.
