#! /usr/bin/env bash
# Copyright 2020 UCLouvain
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


# Main script to evaluate a CHES2020 CTF submission
#
# The evaluation is done in a few stages:
# - unpack:
#   The submission zip file is unpacked and the json description file is validated.
# - setup:
#   The setup script from the submission is run.
# - profile:
#   The profile scripts from the submission are run.
# - attack:
#   The attack scripts from the submission are run.
# - rank:
#   The rank of the correct key is computed from the attack results.
#
# Usage:
# main.sh help
#   to list actions
#
# Environment variables:
# - CTF_SUB_FILE
#   Submission zip file.
# - CTF_BASE_DIR
#   Directory containing all the temporary files to the evaluation.
# - CTF_TRACES_TRAIN_DIR
#   Top-level directory containing the training sets (shall be set to the
#   top-level public set during the envaluation, that is, the directory
#   containing sw3/, sw4/, etc.)
# - CTF_CHALLENGE_DIR
#   Top-level directory containing the traces to attack. Shall have identical
#   structure to the $CTF_TRACES_TRAIN_DIR.
#   The user running this script must have read and write access to all the
#   files in this directory.
# - CTF_CHALLENGE_TMP
#   Temporary directory to store temporary challenge traces. This directory
#   contains hardlinks to files in CTF_CHALLENGE_DIR, hence must be on the same
#   partition.
# - CTF_N_ATTACK_SETS
#   Number of different keys on which the attack should be run (goes from key_O to key.
# - CTF_SUDO_RESOURCES
#   Set to 1 to limit resources (CPU, RAM) usage. Requires sudo rights.
# - CTF_DUMP_PREFIX
#   path where to dump scripts
# - CTF_DUMP_CMD
#   1 to dump to the file

SCRIPT=$(readlink -f $0)
export SCRIPTPATH=`dirname $SCRIPT`

PACKED_SUB_DIR=$CTF_BASE_DIR/packed
PACKED_CTF_SUB_FILE=$PACKED_SUB_DIR/submission.zip
UNPACKED_SUB=$CTF_BASE_DIR/unpacked
SETUP_DIR=$CTF_BASE_DIR/setup
RANKING_DIR=$CTF_BASE_DIR/ranking

DIR_SIZE=$SCRIPTPATH/utils/dir_size.sh

CTF_ATTACK_SETS=$(seq 0 $(($CTF_N_ATTACK_SETS-1)) )

export CTF_SUBMISSION_DIR=$UNPACKED_SUB
export CTF_PROFILE_DIR=$CTF_TRACES_TRAIN_DIR

source $SCRIPTPATH/resource_settings.sh

function get_testplan() {
    TESTPLAN=$($SCRIPTPATH/utils/parse.sh $UNPACKED_SUB)
    if [ "$CTF_DEBUG" != "1" ]
    then
        TESTPLAN=$(echo "$TESTPLAN" | grep -v sw2)
    fi
    echo "$TESTPLAN"
}

function unpack() {
    ## Create dir containing packed submission
    rm -rf $PACKED_SUB_DIR
    mkdir -p $PACKED_SUB_DIR
    # harlink submission
    ln $CTF_SUB_FILE $PACKED_CTF_SUB_FILE

    ## Unpacking submission
    rm -rf $UNPACKED_SUB
    mkdir -p $UNPACKED_SUB
    $SCRIPTPATH/utils/unpack.sh $PACKED_SUB_DIR $UNPACKED_SUB
    rm -rf $PACKED_SUB_DIR

    get_testplan || (echo "Testplan" >&2 && exit -1)
}

function setup() {
    export CTF_EVAL_IMG=$SETUP_DIR/img
    export CTF_DUMP_DEST=${CTF_DUMP_PREFIX}.sh
    $SCRIPTPATH/dump_cmd.sh $SCRIPTPATH/main_setup.sh
}

function profile() {
    export CTF_EVAL_IMG=$SETUP_DIR/img.sif
    get_testplan
    echo "$TESTPLAN" | while read PLAN_CMD
    do
        export CTF_PLAN_TYPE=$(echo $PLAN_CMD | cut -d ' ' -f 1 )
        export CTF_DUMP_DEST=${CTF_DUMP_PREFIX}_${CTF_PLAN_TYPE}.sh
        $SCRIPTPATH/dump_cmd.sh $SCRIPTPATH/main_profile.sh
    done
}

function attack() {
    export CTF_EVAL_IMG=$SETUP_DIR/img.sif
    get_testplan
    echo "$TESTPLAN" | while read PLAN_CMD
    do
        export CTF_PLAN_TYPE=$(echo $PLAN_CMD | cut -d ' ' -f 1 )
        PLAN_CTF_N_TRACES=$(echo $PLAN_CMD | cut -d ' ' -f 2- )
        export CTF_ATTACK_SCRIPT=$CTF_PLAN_TYPE/attack
        export CTF_PROFILE_SCRATCH=$CTF_BASE_DIR/profile/$CTF_PLAN_TYPE/profile_scratch
        for CTF_N_TRACES in $PLAN_CTF_N_TRACES
        do
            export CTF_N_TRACES
            for CTF_ATTACK_SET in $CTF_ATTACK_SETS
            do
                export CTF_ATTACK_SET
                echo "attack $CTF_PLAN_TYPE $CTF_N_TRACES $CTF_ATTACK_SET"
                export CTF_DUMP_DEST=${CTF_DUMP_PREFIX}_${CTF_PLAN_TYPE}_${CTF_N_TRACES}_${CTF_ATTACK_SET}.sh
                $SCRIPTPATH/dump_cmd.sh $SCRIPTPATH/main_attack.sh
            done
        done
    done
}

function rank() {
    get_testplan
    mkdir -p $RANKING_DIR/temp
    echo "$TESTPLAN" | while read PLAN_CMD
    do
        export CTF_PLAN_TYPE=$(echo $PLAN_CMD | cut -d ' ' -f 1 )
        PLAN_CTF_N_TRACES=$(echo $PLAN_CMD | cut -d ' ' -f 2- )
        for CTF_N_TRACES in $PLAN_CTF_N_TRACES
        do
            export CTF_N_TRACES
            for CTF_ATTACK_SET in $CTF_ATTACK_SETS
            do
                export CTF_ATTACK_SET
                export CTF_TEMP_RANKING_DIR=$RANKING_DIR/temp
                export CTF_DUMP_DEST=${CTF_DUMP_PREFIX}_${CTF_PLAN_TYPE}_${CTF_N_TRACES}_${CTF_ATTACK_SET}.sh
                $SCRIPTPATH/dump_cmd.sh $SCRIPTPATH/main_rank.sh
            done
        done
    done
}

function summarize_ranks() {
    get_testplan
    mkdir -p $RANKING_DIR/summary
    echo "$TESTPLAN" | while read PLAN_CMD
    do
        CTF_PLAN_TYPE=$(echo $PLAN_CMD | cut -d ' ' -f 1 )
        PLAN_CTF_N_TRACES=$(echo $PLAN_CMD | cut -d ' ' -f 2- )
        SUM_RANK_FILE=$RANKING_DIR/summary/${CTF_PLAN_TYPE}
        rm -f $SUM_RANK_FILE
        for CTF_N_TRACES in $PLAN_CTF_N_TRACES
        do
            RANKS=()
            for CTF_ATTACK_SET in $CTF_ATTACK_SETS
            do
                RANK_FILE=$RANKING_DIR/temp/${CTF_PLAN_TYPE}_${CTF_N_TRACES}_${CTF_ATTACK_SET}.txt
                RANK=$(<$RANK_FILE)
                RANKS+=($RANK)
            done
            echo -n "$CTF_N_TRACES " >> $SUM_RANK_FILE
            echo $( IFS=";" ; echo "${RANKS[*]}") >> $SUM_RANK_FILE
        done
        $SCRIPTPATH/utils/average_ranks.sh < $SUM_RANK_FILE > ${SUM_RANK_FILE}_avg
    done
}

function all() {
    unpack
    setup
    profile
    attack
    rank
    summarize_ranks
}

# run according to the main command
if [ -z "$1" ]
then
    echo "Give a command to run:"
    echo "  all: run the whole evaluation"
    echo "  unpack: unzip the submission and parse the description file"
    echo "  setup: setup the container and the submission"
    echo "  profile: run the profiling step"
    echo "  attack: run the attack step"
    echo "  rank: rank the correct key according to the results of the attack"
    echo "  summarize_ranks: put the ranks together in a single file and average"
else
    for cmd in $1; do
        $cmd
    done
fi

