#! /usr/bin/env bash
# Copyright 2020 UCLouvain
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# Script for evaluating CHES2020 CTF submissions.
#
# The evaluation is done in a Singularity container. This provides isolation of
# environment and reproducibility.
# The evaluation is done in different stages:
# - cleanup:
#   Remove all temporary and result files.
# - setup:
#   - setup_dirs:
#     Create temporary and result directories
#   - setup_img:
#     Create the singularity sandbox container. This is currently a Debian10 OS.
#     In case usage of another linux OS image is unavoidable, participants are
#     invited to get in touh with the organizers.
#   - setup_submission:
#     Copy the submission file in the container, unzip it and run the "setup"
#     script of the submission package.
#     The intended use for this script is to install the required depenencies,
#     compile software, etc. Access to internet is granted, but no
#     profiling/attack traces are available.
# - profile:
#   The "profile" script of the submission package is run.
#   The intended use-case for the "profile" script is to profile the attack
#   using the supplied traces. In case no profiling is needed (or its result is
#   submitted in the submission package), this script can be left empty.
#   The profiling traces are available read-only at /profile_traces.
#   The /profile_scratch directory is mounted read-write (intended to store the
#   result of the profiling).
# - attack:
#   The "attack" script of the submission package is run.
#   The profiling and attack traces are available read-only at /profile_traces
#   and /challenge_traces respectively.
#   The /profile_scratch is available read-only, and the /attack_scratch is read-write.
#   The attack results should be printed on the standard output.
#
# The script stores execution time (is seconds) for the submission scripts in the
# $CTF_DATA_DIR/reports/time file.
# Their standard output is stored in the $CTF_DATA_DIR/reports/[script].res file.
# The standard error is stored in the $CTF_DATA_DIR/reports/[script].log file.
# The attack script should print its result on its standard output.
#
# Execute this script as:
#   ./run_eval.sh [STAGE]
# or
#   ./run_eval.sh all
# to execute all the stages in the given order.
#
# Note: The resources can be further limited by running this script in a systemd scope:
#  systemd-run --user --scope -p MemoryMax=4G -p CPUQuota=200% ./run_eval.sh [STAGE]
#
# Environment variables are used to parameterize this script. If not provided,
# there is a default value:
# - CTF_CHALLENGE_DIR (default: ./challenge)
#   Directory containing the traces to be attacked.
# - CTF_PROFILE_DIR (default: ./profile)
#   Directory containing the traces to be used for profiling
# - CTF_SUBMISSION_DIR (default: ./submission)
#   Directory containing the submitted scripts.
# - CTF_SETUP_SCRIPT (default: setup)
#   Location of the setup script inside the the CTF_SUBMISSION_DIR.
# - CTF_PROFILE_SCRIPT (default: profile)
#   Location of the profile script inside the the CTF_SUBMISSION_DIR
# - CTF_ATTACK_SCRIPT (default: attack)
#   Location of the attack script inside the the CTF_SUBMISSION_DIR
# - CTF_PROFILE_SCRATCH (default: $CTF_DATA_DIR/profile_scratch)
#   Location of the scratch directory for the profile phase.
# - CTF_ATTACK_SCRATCH (default: $CTF_DATA_DIR/attack_scratch)
#   Location of the scratch directory for the attack phase.
# - CTF_EVAL_IMG (default: $CTF_DATA_DIR/img)
#   Location of the Singuarity sandbox
# - CTF_TIMEOUT_SETUP (default: 10m)
#   Max execution time for the setup_submission phase
# - CTF_TIMEOUT_PROFILE (default: 1h)
#   Max execution time for the profile phase
# - CTF_TIMEOUT_ATTACK (default: 20m)
#   Max execution time for the attack phase
# - CTF_DATA_DIR (default: ./ctf_eval)
#   Directory used for storing the container, scratch directories and result files
# - CTF_DEBIAN_MIRROR (default: deb.debian.org)
#   Debian mirror to be used in the container
#    
# SEE ALSO
# The Submission Format documentation for more information on the exepected
# format of the submission file, the format of the profile and attack traces,
# and the expected result of the attack script.
#
# DEPENDENCIES
# This script uses Singularity containers in fakeroot mode and has been tested
# with Singularity 3.5.
# See https://sylabs.io/guides/3.5/user-guide/

CTF_CHALLENGE_DIR=${CTF_CHALLENGE_DIR:-./challenge}
CTF_CHALLENGE_DIR_TRACES=${CTF_CHALLENGE_DIR_TRACES:-./challenge}
CTF_PROFILE_DIR=${CTF_PROFILE_DIR:-./profile}
CTF_SUBMISSION_DIR=${CTF_SUBMISSION_DIR:-./submission}
CTF_SETUP_SCRIPT=${CTF_SETUP_SCRIPT:-setup}
CTF_PROFILE_SCRIPT=${CTF_PROFILE_SCRIPT:-profile}
CTF_ATTACK_SCRIPT=${CTF_ATTACK_SCRIPT:-attack}

# Debian mirror to be used in the container
CTF_DEBIAN_MIRROR=${CTF_DEBIAN_MIRROR:-deb.debian.org}

# Max execution time for the setup phase
CTF_TIMEOUT_SETUP=${CTF_TIMEOUT_SETUP:-10m}
# Max execution time for the profiling phase
CTF_TIMEOUT_PROFILE=${CTF_TIMEOUT_PROFILE:-1h}
# Max execution time for the attack phase
CTF_TIMEOUT_ATTACK=${CTF_TIMEOUT_ATTACK:-20m}

# directory where to put the reports and the scratch
CTF_DATA_DIR=${CTF_DATA_DIR:-./ctf_eval}

CTF_PROFILE_SCRATCH=${CTF_PROFILE_SCRATCH:-$CTF_DATA_DIR/profile_scratch}
CTF_ATTACK_SCRATCH=${CTF_ATTACK_SCRATCH:-$CTF_DATA_DIR/attack_scratch}
REPORTS_DIR=$CTF_DATA_DIR/reports

CTF_EVAL_IMG=${CTF_EVAL_IMG:-$CTF_DATA_DIR/img}

TMP_DIR=$CTF_DATA_DIR/tmp

TMP_BIND=$TMP_DIR:/tmp:rw
PROF_BIND=$CTF_PROFILE_DIR:/profile_traces
CTF_PROFILE_SCRATCH_BIND=$CTF_PROFILE_SCRATCH:/profile_scratch
ATTACK_BIND=$CTF_CHALLENGE_DIR:/challenge_traces:ro,$CTF_CHALLENGE_DIR_TRACES:/challenge_dir_traces:ro
CTF_ATTACK_SCRATCH_BIND=$CTF_ATTACK_SCRATCH:/attack_scratch
SUBMISSION_BIND=$CTF_SUBMISSION_DIR:/submission:ro

SINGULARITY_OPTIONS_COMMON="--no-home --containall --hostname ctf --bind $TMP_BIND"
SINGULARITY_OPTIONS_SETUP="$SINGULARITY_OPTIONS_COMMON --fakeroot --writable --bind $SUBMISSION_BIND"
SINGULARITY_OPTIONS_PROFILE="$SINGULARITY_OPTIONS_COMMON --net --network none --bind $PROF_BIND:ro,$CTF_PROFILE_SCRATCH_BIND:rw,$SUBMISSION_BIND"
SINGULARITY_OPTIONS_ATTACK="$SINGULARITY_OPTIONS_COMMON --net --network none --bind $PROF_BIND:ro,$CTF_PROFILE_SCRATCH_BIND:ro,$ATTACK_BIND,$CTF_ATTACK_SCRATCH_BIND:rw,$SUBMISSION_BIND"

function cleanup_temp() {
    rm -rf $TMP_DIR
    mkdir -p $TMP_DIR
}

function cleanup() {
    rm -rf $CTF_PROFILE_SCRATCH $CTF_ATTACK_SCRATCH $REPORTS_DIR $TMP_DIR
    rm -rf $CTF_EVAL_IMG
}

function setup_dirs() {
    mkdir -p $CTF_PROFILE_SCRATCH $CTF_ATTACK_SCRATCH $REPORTS_DIR
}

function setup_img() {
    cleanup_temp
    singularity build --sandbox $CTF_EVAL_IMG library://library/default/debian:10
    timeout -s KILL 2m \
        singularity shell $SINGULARITY_OPTIONS_COMMON --writable $CTF_EVAL_IMG \
        <<EOF
chmod 1777 /tmp
sed -i -e "s/ftp.us.debian.org/$CTF_DEBIAN_MIRROR/" /etc/apt/sources.list
apt-get --yes update
mkdir -p /submission /challenge_traces /challenge_dir_traces /profile_traces /profile_scratch /attack_scratch
EOF
}

function setup_submission() {
    echo "setup submission"
    cleanup_temp
    /usr/bin/time -a -o $REPORTS_DIR/time -f "setup %e" \
        timeout -s KILL --foreground $CTF_TIMEOUT_SETUP \
        sh -c "singularity exec $SINGULARITY_OPTIONS_SETUP $CTF_EVAL_IMG /submission/$CTF_SETUP_SCRIPT > $REPORTS_DIR/setup.res 2> $REPORTS_DIR/setup.log"
}

function setup() {
    setup_dirs
    setup_img
    setup_submission
}

function package_image() {
    singularity build --fakeroot $1 $CTF_EVAL_IMG
}

function profile() {
    echo "#profile"
    setup_dirs
    cleanup_temp
    /usr/bin/time -a -o $REPORTS_DIR/time -f "profile %e" \
        timeout -s KILL $CTF_TIMEOUT_PROFILE \
        singularity exec $SINGULARITY_OPTIONS_PROFILE $CTF_EVAL_IMG /submission/$CTF_PROFILE_SCRIPT \
        > $REPORTS_DIR/profile.res 2> $REPORTS_DIR/profile.log
}

function attack() {
    echo "#attack"
    setup_dirs
    cleanup_temp
    /usr/bin/time -a -o $REPORTS_DIR/time -f "attack %e" \
        timeout -s KILL $CTF_TIMEOUT_ATTACK \
        singularity exec $SINGULARITY_OPTIONS_ATTACK $CTF_EVAL_IMG /submission/$CTF_ATTACK_SCRIPT \
        > $REPORTS_DIR/attack.res 2> $REPORTS_DIR/attack.log
}

function shell() {
    singularity shell $SINGULARITY_OPTIONS_SETUP --bind $PROF_BIND:ro,$CTF_PROFILE_SCRATCH_BIND:rw,$ATTACK_BIND:ro,$CTF_ATTACK_SCRATCH_BIND:rw $CTF_EVAL_IMG /submission/$CTF_ATTACK_SCRIPT
}

function all() {
    cleanup
    setup
    profile
    attack
}

# run according to the main command
if [ -z "$1" ]
then
    echo "Give a command to run:"
    echo "  all: run the whole evaluation"
    echo "  setup: setup the container and the submission"
    echo "  profile: run the profiling step"
    echo "  attack: run the attack step"
    echo "  package_image [IMG]: pack the image to SIF format"
else
    $1 $2
fi
