#! /usr/bin/env python3
# Copyright 2020 UCLouvain
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# Parse a submission.json file.

import sys
import json
import os

import jsonschema
import spdx_license_list

SUB_JSON='submission.json'

submission_root = sys.argv[1]

CONTESTS = ['sw2', 'sw3', 'sw4', 'sw6', 'sw8', 'hw2', 'hw3', 'hw4']
MAX_CONTEST_TRACES = {
        'sw2': 100000-1,
        'sw3': 100000-1,
        'sw4': 100000-1,
        'sw6': 100000-1,
        'sw8': 100000-1,
        'hw2': 5000000-1,
        'hw3': 5000000-1,
        'hw4': 5000000-1
        }

SCHEMA = {
"$schema": "http://json-schema.org/schema#",
"type": "object",
"properties": {
    "authors": {
        "type": "array",
        "items": {
            "type": "object",
            "properties": {
                "name": { "type": "string" },
                "email": {
                    "type": "string",
                    "format": "email"
                    }
                },
            "required": ["name", "email"]
            },
        "minItems": 1,
        "uniqueItems": True
        },
    "name": { "type": "string" },
    "license": { "type": "string" },
    "attacks": {
        "type": "object",
        "patternProperties": {
            "^(sw|hw)[0-9]+$": {
                "type": "object",
                "properties": {
                    "n_traces": {
                        "type": "array",
                        "items": { "type": "integer" },
                        "minItems": 1,
                        "maxItems": 8,
                        }
                    },
                "required": ["n_traces"]
                },
            }
        },
    },
    "required": ["authors", "name", "license", "attacks"],
}


with open(os.path.join(submission_root, SUB_JSON)) as f:
    submission = json.load(f)

jsonschema.validate(instance=submission, schema=SCHEMA)
license = submission["license"]

if license not in spdx_license_list.LICENSES:
    raise ValueError("LICENSE '{}' is not a SPDX license identifier. See https://spdx.org/licenses/ for the list of valid identifiers".format(license))
if not spdx_license_list.LICENSES[license]["isOsiApproved"]:
    raise ValueError("LICENSE {} is not OSI-approved. See https://spdx.org/licenses/ for the list of OSI-approved licenses.".format(license))

for contest in CONTESTS:
    if contest in submission['attacks']:
        n_traces = submission['attacks'][contest]['n_traces']
        for n_trace in n_traces:
            if n_trace <= 0:
                raise ValueError("Number of traces {} is not positive.".format(n_trace))
            if n_trace > MAX_CONTEST_TRACES[contest]:
                raise ValueError("Number of traces {} is not multiple of 100.".format(n_trace))
        print('{} {}'.format(contest, ' '.join(map(str, n_traces))))


